import { ComputedRef } from 'vue'
import { Plugin as Plugin_2 } from 'vue'
import { reactive } from 'vue'
import { Ref } from 'vue'
import { ref } from 'vue'
import { UnwrapRef } from 'vue'

export declare type Configuration =
  keyof UseValidation_CustomValidationBehaviorFunctions extends never
    ? {
        defaultValidationBehavior: ValidationBehaviorString
        validationBehavior?: ConfigurationValidationBehavior
      }
    : {
        defaultValidationBehavior: ValidationBehaviorString
        validationBehavior: ConfigurationValidationBehavior
      }

export declare type ConfigurationValidationBehavior = n_domain.Optional<
  {
    [K in ValidationBehaviorString]: ValidationBehaviorFunction
  },
  DefaultValidationBehaviorString
>

/**
 *
 * Configure the validation behavior of `useValidation`.
 *
 * @param configuration - The Form validation configuration
 */
export declare function createValidation(configuration: Configuration): Plugin_2

declare function debounce<TArgs extends unknown[]>(
  target: (...args: [...TArgs]) => void,
  {
    wait,
    shouldInvoke
  }: {
    wait: number
    shouldInvoke?: (...args: [...TArgs]) => ShouldInvokeResult
  }
): (...args: [...TArgs]) => void

declare function deepCopy<T>(toClone: T): T

declare type DeepIndex<T, Ks extends readonly Key[], R = unknown> = Ks extends [
  infer First,
  ...infer Rest
]
  ? First extends keyof T
    ? Rest extends readonly Key[]
      ? DeepIndex<T[First], Rest>
      : R
    : R
  : T

declare function deepIterator(
  obj: object,
  predicate?: (value: object) => boolean
): Generator<DeepIteratorResult, void>

declare type DeepIteratorResult = {
  key: string
  value: any
  parent: any
  path: string[]
  isLeaf: boolean
}

declare type DeepMaybeRef<T> = T extends Ref
  ? T | UnwrapRef<T>
  : T extends Record<string, unknown>
  ? DeepMaybeRefRecord<T>
  : n_domain.MaybeRef<T>

declare type DeepMaybeRefRecord<
  T extends Record<n_domain.Key, unknown> | undefined
> = T extends undefined
  ? undefined
  : {
      [K in keyof T]: T[K] extends Ref
        ? T[K] | UnwrapRef<T[K]>
        : T[K] extends any[]
        ? n_domain.MaybeRef<T[K]>
        : T[K] extends Record<string, unknown> | undefined
        ? DeepMaybeRefRecord<T[K]>
        : n_domain.MaybeRef<T[K]>
    }

declare type DefaultValidationBehaviorString =
  | 'aggressive'
  | 'lazy'
  | 'lazier'
  | 'submit'

declare function disposeForm(form: Form, deletedFormData: any): void

export declare type Field<
  TValue,
  TExtra extends Record<n_domain.Key, unknown> = Record<string, never>
> = {
  /**
   *
   * The field's default value.
   */
  $value: DeepMaybeRef<TValue>
  /**
   *
   * Rules to use for validation.
   */
  $rules?: FieldRule<TValue>[]
} & (TExtra extends Record<string, never> ? unknown : TExtra)

export declare type FieldNames<T> = T extends (infer TArray)[]
  ? FieldNames<TArray>
  : {
      [K in keyof T]-?: T[K] extends
        | {
            $value: any
          }
        | undefined
        ? K
        : FieldNames<T[K]>
    }[keyof T]

declare type FieldRule<
  TSimpleParameter,
  TKeyedParameters extends readonly any[] = any[]
> =
  | FieldSimpleRule<
      TSimpleParameter extends any[]
        ? TSimpleParameter
        : UnwrapRef<TSimpleParameter>
    >
  | FieldRuleWithKey<TKeyedParameters>

declare type FieldRuleWithKey<TParameters extends readonly any[]> =
  | RuleWithKey<TParameters>
  | [
      validationBehavior: ValidationBehavior,
      rule: RuleWithKey<TParameters>,
      debounce?: number
    ]
  | [rule: RuleWithKey<TParameters>, debounce: number]

declare type FieldSimpleRule<TParameter = any> =
  | SimpleRule<TParameter>
  | [
      validationBehavior: ValidationBehavior,
      rule: SimpleRule<TParameter>,
      debounce?: number
    ]
  | [rule: SimpleRule<TParameter>, debounce: number]

declare class Form {
  private _simpleValidators
  private _keyedValidators
  private _reactiveFieldMap
  tryGetSimpleValidators: ({
    success,
    failure
  }: {
    success(value: SimpleValidators): void
    failure?(): void
  }) => (key: number) => void
  trySetKeyedValidators: ({
    success,
    failure
  }: {
    success?(value: KeyedValidators): void
    failure?(value: KeyedValidators): void
  }) => (key: string, value: KeyedValidators) => void
  tryGetKeyedValidators: ({
    success,
    failure
  }: {
    success(value: KeyedValidators): void
    failure?(): void
  }) => (key: string) => void
  rulesValidating: Ref<number>
  submitting: Ref<boolean>
  validating: ComputedRef<boolean>
  hasError: ComputedRef<boolean>
  errors: ComputedRef<string[]>
  registerField(
    uid: number,
    name: string,
    modelValue: unknown,
    ruleInfos: RuleInformation[]
  ): FormField
  validate(uid: number, force?: boolean): void
  validateAll(names?: readonly n_domain.Key[]): Promise<void>
  dispose(uid: number): void
  resetFields(): void
  getField(uid: number): FormField | undefined
  private _invokeValidatorsForKeys
  private _collectValidatorResultsForKeys
  private _collectValidatorResultsForNames
  private _isEveryFieldTouched
}

declare class FormField {
  private _rules
  private _validationBehaviors
  private _buffers
  private _watchStopHandle
  private _form
  uid: number
  rulesValidating: Ref<number>
  initialModelValue: unknown
  name: string
  touched: Ref<boolean>
  dirty: Ref<boolean>
  modelValue: ReturnType<typeof ref> | ReturnType<typeof reactive>
  rawErrors: (string | null)[]
  errors: ComputedRef<string[]>
  validating: ComputedRef<boolean>
  hasError: ComputedRef<boolean>
  constructor(
    form: Form,
    uid: number,
    name: string,
    modelValue: any,
    ruleInfos: RuleInformation[]
  )
  validate(
    ruleNumber: number,
    modelValues: unknown[],
    noThrow: boolean
  ): Promise<void>
  reset(resetValue?: unknown): void
  dispose(): void
  shouldValidate(ruleNumber: number, force: boolean, submit: boolean): any
  private _setError
}

declare function getResultFormData(
  transformedFormData: any,
  predicate?: (
    value: Omit<n_domain.DeepIteratorResult, 'isLeaf' | 'parent'>
  ) => unknown
): any

declare const isArray: (x: unknown) => x is any[]

declare const isDefined: <T>(x: T | null | undefined) => x is T

declare const isField: <T>(x: unknown) => x is {
  /**
   *
   * The field's default value.
   */
  $value: DeepMaybeRef<T>
  /**
   *
   * Rules to use for validation.
   */
  $rules?: FieldRule<T, any[]>[] | undefined
}

declare const isObject: (x: unknown) => x is Record<Key, any>

declare const isRecord: (x: unknown) => x is Record<Key, any>

declare const isSimpleRule: (
  rule: SimpleRule | RuleWithKey
) => rule is SimpleRule<any>

declare const isTransformedField: <T>(x: unknown) => x is {
  /**
   *
   * The unique id of this field.
   */
  $uid: number
  /**
   *
   * The current field's value.
   */
  $value: T
  /**
   *
   * A list of validation error messages local to the field without `null` values.
   */
  $errors: string[]
  /**
   *
   * The field's raw error messages one for each rule and `null` if there is no error.
   */
  $rawErrors: (string | null)[]
  /**
   *
   * `True` while there are any errors on the field.
   */
  $hasError: boolean
  /**
   *
   * `True` while the field has any pending rules.
   */
  $validating: boolean
  /**
   *
   * `True` if the field is touched.
   *
   * @remarks
   * In most cases, it should be set together with the `blur` event.
   * Either through `$validate` or manually.
   */
  $touched: boolean
  /**
   *
   * `True` if the `$value` of the field has changed at least once.
   */
  $dirty: boolean
  /**
   *
   * Validate the field.
   *
   * @param options - Validation options to use
   * @default
   * ```
   * { setTouched: true, force: true }
   * ```
   */
  $validate(options?: ValidateOptions | undefined): void
}

declare type Key = string | number

export declare type KeyedRule<TParameters extends readonly any[] = any[]> = (
  ...values: [...TParameters]
) => any

declare type KeyedValidator = {
  validator: Validator
  validatorNotDebounced: Validator
  meta: {
    field: FormField
  }
}

declare type KeyedValidators = Set<KeyedValidator>

declare class LinkedList<T> {
  first: LinkedListNode<T> | null
  last: LinkedListNode<T> | null
  count: number
  addFirst(value: T): LinkedListNode<T>
  addLast(value: T): LinkedListNode<T>
  remove(node: LinkedListNode<T>): void
  removeFirst(): void
  removeLast(): void
  nodesForwards(): Generator<LinkedListNode<T>, void, unknown>
  nodesBackwards(): Generator<LinkedListNode<T>, void, unknown>
}

declare class LinkedListNode<T> {
  value: T
  next: LinkedListNode<T> | null
  prev: LinkedListNode<T> | null
  constructor(value: T)
}

declare type MaybeRef<T> = Ref<T> | T

declare namespace n_domain {
  export {
    deepCopy,
    tryGet,
    trySet,
    path,
    set,
    uid,
    deepIterator,
    DeepIteratorResult,
    debounce,
    isArray,
    isDefined,
    isObject,
    isRecord,
    LinkedList,
    PromiseCancel,
    Key,
    MaybeRef,
    DeepIndex,
    Tuple,
    Optional
  }
}

declare namespace n_form {
  export {
    disposeForm,
    getResultFormData,
    resetFields,
    transformFormData,
    isField,
    isTransformedField,
    Form,
    FormField,
    ValidationError,
    isSimpleRule,
    FieldRule,
    KeyedRule,
    SimpleRule,
    RuleWithKey,
    ValidationBehavior,
    ValidationBehaviorFunction,
    DefaultValidationBehaviorString,
    ValidationBehaviorInfo,
    ValidationBehaviorString,
    Field,
    TransformedField,
    TransformedFormData,
    FieldNames,
    ResultFormData,
    DeepMaybeRef
  }
}

declare type Optional<T, K extends keyof T> = Partial<Pick<T, K>> & Omit<T, K>

declare function path(path: readonly Key[], obj: Record<Key, unknown>): any

declare class PromiseCancel<T = unknown> {
  private promise
  private resolve
  private reject
  constructor()
  cancelResolve(value: T | PromiseLike<T>): void
  cancelReject(reason?: any): void
  race<Ps extends readonly Promise<any>[]>(
    ...promises: [...Ps]
  ): Promise<
    T | ([...Ps][number] extends PromiseLike<infer U> ? U : [...Ps][number])
  >
  private assign
}

declare function resetFields(
  form: Form,
  data: any,
  transformedFormData: any
): void

declare type ResultFormData<FormData extends object | undefined> =
  FormData extends undefined
    ? undefined
    : {
        [K in keyof FormData]: FormData[K] extends
          | {
              $value: infer TValue
            }
          | undefined
          ? UnwrapRef<TValue>
          : FormData[K] extends object | undefined
          ? ResultFormData<FormData[K]>
          : FormData[K]
      }

declare type RuleInformation = {
  validationBehavior: ValidationBehaviorFunction
  rule: SimpleRule | RuleWithKey
  debounce?: number
}

declare type RuleWithKey<T extends readonly any[] = any[]> = {
  key: string
  rule?: KeyedRule<T>
}

declare function set(obj: any, keys: readonly Key[], value: any): void

declare type ShouldInvokeResult = boolean | void

export declare type SimpleRule<TParameter = any> = (
  ...value: TParameter[]
) => any

declare type SimpleValidators = {
  validators: Validator[]
  validatorsNotDebounced: Validator[]
  meta: {
    field: FormField
    keys: string[]
    rollbacks: (() => void)[]
  }
}

export declare type TransformedField<
  TValue,
  TExtra extends Record<n_domain.Key, unknown> = Record<string, never>
> = {
  /**
   *
   * The unique id of this field.
   */
  $uid: number
  /**
   *
   * The current field's value.
   */
  $value: TValue
  /**
   *
   * A list of validation error messages local to the field without `null` values.
   */
  $errors: string[]
  /**
   *
   * The field's raw error messages one for each rule and `null` if there is no error.
   */
  $rawErrors: (string | null)[]
  /**
   *
   * `True` while there are any errors on the field.
   */
  $hasError: boolean
  /**
   *
   * `True` while the field has any pending rules.
   */
  $validating: boolean
  /**
   *
   * `True` if the field is touched.
   *
   * @remarks
   * In most cases, it should be set together with the `blur` event.
   * Either through `$validate` or manually.
   */
  $touched: boolean
  /**
   *
   * `True` if the `$value` of the field has changed at least once.
   */
  $dirty: boolean
  /**
   *
   * Validate the field.
   *
   * @param options - Validation options to use
   * @default
   * ```
   * { setTouched: true, force: true }
   * ```
   */
  $validate(options?: ValidateOptions): void
} & (TExtra extends Record<string, never> ? unknown : UnwrapRef<TExtra>)

declare type TransformedFormData<FormData extends object | undefined> =
  FormData extends undefined
    ? undefined
    : {
        [K in keyof FormData]: FormData[K] extends
          | {
              $value: infer TValue
            }
          | undefined
          ? FormData[K] extends undefined
            ? undefined
            : TransformedField<
                UnwrapRef<TValue>,
                Omit<Exclude<FormData[K], undefined>, '$value' | '$rules'>
              >
          : FormData[K] extends object | undefined
          ? TransformedFormData<FormData[K]>
          : FormData[K]
      }

declare function transformFormData(form: Form, formData: object): void

declare const tryGet: <K, V>(
  map: Map<K, V>
) => ({
  success,
  failure
}: {
  success(value: V): void
  failure?(): void
}) => (key: K) => void

declare const trySet: <K, V>(
  map: Map<K, V>
) => ({
  success,
  failure
}: {
  success?(value: V): void
  failure?(value: V): void
}) => (key: K, value: V) => void

declare type Tuple<T, N extends number> = number extends N
  ? T[]
  : _Tuple<T, N, []>

declare type _Tuple<
  T,
  N extends number,
  R extends unknown[]
> = R['length'] extends N ? R : _Tuple<T, N, [T, ...R]>

declare function uid(): number

export declare type UseValidation<FormData extends object> = {
  /**
   *
   * A transformed reactive `formData` object.
   */
  form: n_form.TransformedFormData<FormData>
  /**
   *
   * `True` during validation after calling `validateFields` when there were rules that returned a `Promise`.
   */
  submitting: Ref<boolean>
  /**
   *
   * `True` while the form has any pending rules.
   */
  validating: ComputedRef<boolean>
  /**
   *
   * `True` if the form has any error.
   */
  hasError: ComputedRef<boolean>
  /**
   *
   * All current validation error messages.
   */
  errors: ComputedRef<string[]>
  /**
   *
   * Validate all fields and return a `Promise` containing the resulting form data.
   *
   * @param options - Options to use for validation
   * @throws `ValidationError`
   */
  validateFields(options?: {
    /**
     *
     * A list of field names to validate.
     *
     * @default
     * ```
     * undefined // meaning validate all
     * ```
     */
    names?: n_form.FieldNames<FormData>[] | string[]
    /**
     *
     * Filter which values to keep in the resulting form data.
     *
     * @remarks
     * Used like `Array.prototype.filter`.
     *
     * @default
     * ```
     * () => true // meaning keep all
     * ```
     */
    predicate?: (
      value: Omit<n_domain.DeepIteratorResult, 'isLeaf' | 'parent'>
    ) => unknown
  }): Promise<n_form.ResultFormData<FormData>>
  /**
   *
   * Reset all fields to their default value or pass an object to set specific values.
   *
   * @remarks
   * It will not create any new fields that are not present in the form data initially.
   *
   * @param formData - `FormData` to set specific values. It has the same structure as the object passed to `useValidation`.
   */
  resetFields(formData?: Partial<n_form.ResultFormData<FormData>>): void
  /**
   *
   * Adds a new property to the form data.
   *
   * @remarks
   * Fields with a `$value` are transformed.
   *
   * @param path - A path of `string` and `numbers`
   * @param value - The value to add at the specified path
   */
  add<Ks extends readonly n_domain.Key[]>(
    path: readonly [...Ks],
    value: n_domain.DeepIndex<FormData, Ks> extends (infer TArray)[]
      ? TArray
      : n_domain.DeepIndex<FormData, Ks>
  ): void
  /**
   *
   * Removes a property from the form data.
   *
   * @param path - A path of `string` and `numbers` to the property to remove
   */
  remove(path: n_domain.Key[]): void
}

/**
 *
 * Vue composition function for form validation.
 *
 * @remarks
 * For type inference inside of `useValidation` you need to define the structure of your
 * `formData` upfront and pass it as the generic parameter `FormData`.
 *
 * @param formData - The structure of your `formData`.
 *
 * @example
 * ```
 * type FormData = {
 *   name: Field<string>,
 *   password: Field<string>
 * }
 *
 * const { form } = useValidation<FormData>({})
 * ```
 */
export declare function useValidation<FormData extends object>(
  formData: FormData
): UseValidation<FormData>

declare type ValidateOptions = {
  /**
   *
   * Set the field touched when called.
   *
   * @default true
   */
  setTouched?: boolean
  /**
   *
   * Validate with the `force` flag set.
   *
   * @default true
   */
  force?: boolean
}

export declare type ValidationBehavior =
  | ValidationBehaviorString
  | ValidationBehaviorFunction

export declare type ValidationBehaviorFunction = (
  info: ValidationBehaviorInfo
) => any

export declare type ValidationBehaviorInfo<T = any> = {
  /**
   *
   * `True` if the paired rule of this behavior has an error.
   */
  hasError: boolean
  /**
   *
   * The touched state of the field.
   */
  touched: boolean
  /**
   *
   * The dirty state of the field.
   */
  dirty: boolean
  /**
   *
   * `True` if the validation was triggered with the `force` flag.
   */
  force: boolean
  /**
   *
   * `True` if the validation was triggered with the `submit` flag.
   */
  submit: boolean
  /**
   *
   * The `$value` of the field.
   */
  value: T
}

export declare type ValidationBehaviorString =
  | DefaultValidationBehaviorString
  | keyof UseValidation_CustomValidationBehaviorFunctions

export declare class ValidationError extends Error {
  constructor()
}

declare type Validator = (
  modelValues: unknown[],
  force: boolean,
  submit: boolean
) => ValidatorResult

declare type ValidatorResult = Promise<void | string> | void

export {}

declare global {
  // eslint-disable-next-line @typescript-eslint/no-empty-interface
  interface UseValidation_CustomValidationBehaviorFunctions {}
}
